/* Copyright(c) 2002,2003 Device Drivers Limited, All rights reserved. */
/*                              info@devdrv.com http://www.devdrv.com/ */
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>   /* printk() */
#include <linux/fs.h>       /* everything... */
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#define BUFSIZ 128
static char st_b[BUFSIZ] = "0\n"; /* data buffer, initialize value is "0" */

/*
 * hello_read_proc -- called when user is reading
 */
int hello_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
    int data_length;
    int return_length;

    printk("read_proc() = %s, count = %d, off = %ld\n",
	   st_b, count, offset);

    data_length = strlen(st_b);
    if ((offset + count) > data_length) {
        *eof = 1;
		if ((return_length = data_length - offset) <= 0) {
	    	return 0; /* for read after EOF and 0 byte read */
		}
    }
    else {
        return_length = count;
    }
    strncpy(buf, st_b + offset, return_length);

    *start = buf + offset; /* update for next start point */

    return return_length;
}

/*
 * hello_read_proc -- called when user is writing
 */
int hello_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
    int length;

    length = count > (BUFSIZ - 1) ? (BUFSIZ - 1) : count; /* max bytes limit */

    copy_from_user(st_b, buf, length);
    st_b[length] = '\0';

    printk("write_hello(), buf = %s, count = %lu, length = %d\n",
	   st_b, count, length);

    return length;
}

static void hello_create_proc(void)
{
    struct proc_dir_entry *entry;
    entry = create_proc_entry("hello", 0, 0); /* register "hello" procfile */
    entry->read_proc = hello_read_proc; /* register read routine*/
    entry->write_proc = hello_write_proc; /* register write routine */
}

static void hello_remove_proc(void)
{
    remove_proc_entry("hello", NULL /* parent dir */);
}

static int __init hello_init_module(void)
{
    hello_create_proc();
    printk("<1>Hello, world data = %s\n", st_b);
    return 0;
}

static void __exit hello_cleanup_module(void)
{
    printk("<1>Goodbye world data = %s\n", st_b);
    hello_remove_proc();
}

MODULE_DESCRIPTION("Sample Hello Module");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Atomu Hidaka <hidaka@devdrv.com>");
MODULE_PARM(debug,       "i");
MODULE_PARM_DESC(debug, "hello debug level (0-6)");

module_init(hello_init_module);
module_exit(hello_cleanup_module);
